//	EscWebServer.c
//	This is some minimalist HTTP server functionality, for demonstration at
//	the 2008 Embedded Systems Conference in Boston.

#include "stdinc.h"
#include "EscWeb.h"

char * GetErrorMessage(int nErr, bool * pbRetry);
int GetLastError(SOCKET nSock);
void SysTickleWatchdogs(void);

bool bKillWebThread;

static int nSockWeb;
static char sPageBuff[HTTP_SEND_BUFF_SIZE];
char szOK[] = "OK";
static char szNotFound[] = "Not Found";
static char szNotImpl[] = "Not Implemented";
char szTextHtml[] = "Content-type: text/html\r\n";
char szTextPlain[] = "Content-type: text/plain\r\n";
static char szHeaderEnd[] = "\r\n\r\n";

//	BindSocketPort
//	Helper function for setting up a listen socket on a particular port
static bool BindSocketPort(SOCKET nSock, WORD nPort)
{	bool b;
	struct sockaddr_in sin;
	int n;
	char buff[16];

	memset(&sin, 0, sizeof(sin));
	sprintf(buff, "%d.%d.%d.%d", (nNetworkIP >> 24) & 0xFF, (nNetworkIP >> 16)
		& 0xFF, (nNetworkIP >> 8) & 0xFF, nNetworkIP & 0xFF);
	sin.sin_addr.s_addr = inet_addr(buff);//htonl(0);
	sin.sin_family = PF_INET;
	sin.sin_port = htons(nPort);
	n = bind(nSock, (struct sockaddr*)&sin, sizeof(sin));
	if (n == SOCKET_ERROR)
	{	n = GetLastError(nSock);
		dprintf(PRINTF_ERROR|PRINTF_FLASH, GetErrorMessage(n, &b));
		dprintf(PRINTF_ERROR|PRINTF_FLASH, " - BindSocketPort\n");
		return false;
	}
	return true;
}

//	ListenSock
//	Helper function for starting a listen on a socket
static bool ListenSock(SOCKET nSock)
{	bool b;
	int n = listen(nSock, 1);
	if (n == SOCKET_ERROR)
	{	n = GetLastError(nSock);
		dprintf(PRINTF_ERROR|PRINTF_FLASH, GetErrorMessage(n, &b));
		dprintf(PRINTF_ERROR|PRINTF_FLASH, " - ListenSock\n");
		return false;
	}
	return true;
}

//	SendTcp
//	This sends a packet of chars through the connected socket.
void SendTcp(SOCKET nSock, const char * pBuff, int cbBuff)
{	int n, cb = 0, cbSend;
	while (cb < cbBuff)
	{	cbSend = cbBuff - cb;
		n = send(nSock, (char *)pBuff+cb, cbSend, 0);
		if (n == SOCKET_ERROR)
		{	bool b;
			dprintf(PRINTF_ERROR|PRINTF_FLASH, GetErrorMessage(
				GetLastError(nSock), &b));
			dprintf(PRINTF_ERROR|PRINTF_FLASH, " - SendTcp\n");
			break;
		}
		if (! n)
			break;
		SysYield();
		SysTickleWatchdogs();
		cb += n;
	}
}

//	HttpResponse
//	This formulates a reply to an HTTP request.
void HttpResponse(SOCKET nSock, int nRetCode, const char * pszRetMsg,
	const char * pszHeader, const char * pBody, int cbBody, bool bSendBody,
	bool bKeepAlive)
{	char hbuff[HTTP_HEADER_SIZE];
	int cb;
	cb = sprintf(hbuff, "HTTP/1.0 %u %s\r\n", nRetCode, pszRetMsg);
	if (pszHeader)
		cb += sprintf(hbuff+cb, "%s", pszHeader);
	if (bKeepAlive)
		cb += sprintf(hbuff+cb, "Connection: Keep-Alive\r\n");
	if (cbBody)
		cb += sprintf(hbuff+cb, "Content-Length: %u\r\n", cbBody);
	cb += sprintf(hbuff+cb, "\r\n");
	SendTcp(nSock, hbuff, cb);
	if (cbBody && bSendBody)
		SendTcp(nSock, pBody, cbBody);
}

//	HttpError
//	This handles simple error response cases.
void HttpError(SOCKET nSock, int nRetCode, const char * pszRetMsg)
{	char ebuff[256];
	sprintf(ebuff, "<html><body><h1>Error %u</h1><p><b>%s</b></p></body>"
		"</html>", nRetCode, pszRetMsg);
	HttpResponse(nSock, nRetCode, pszRetMsg, szTextHtml, ebuff, strlen(
		ebuff), true, false);
}

//	HttpHtmlResponse
//	This is the standard case of a successful page return.
void HttpHtmlResponse(SOCKET nSock, struct EscWebRequest * pReq, const char
	* pEnd)
{	HttpResponse(nSock, 200, szOK, szTextHtml, pReq->m_pBuff, pEnd-
		(pReq->m_pBuff), pReq->m_bSendBody, pReq->m_bKeepAlive);
}

//	HttpDispatchRequest
//	This processes a parsed request.  We search our list of page handlers
//	and our list of static resources for something to handle this.
static void HttpDispatchRequest(SOCKET nSock, struct EscWebRequest * pReq)
{	int i;
	//	check for page handlers...
	for (i=0; tblWebPageHandler[i].m_pszPath; i++)
		if (_stricmp(pReq->m_pszPath, tblWebPageHandler[i].m_pszPath) == 0)
		{	tblWebPageHandler[i].m_pfn(nSock, pReq);
			return;
		}
	//	check for static resources...
	for (i=0; tblWebResource[i].m_pszPath; i++)
		if (_stricmp(tblWebResource[i].m_pszPath, pReq->m_pszPath) == 0)
		{	int cb;
			const char * p = (const char *)(tblWebResource[i].m_pData);
			if (tblWebResource[i].m_cbData == -1)
				cb = strlen(p);
			else
				cb = tblWebResource[i].m_cbData;
			HttpResponse(nSock, 200, szOK, tblWebResource[i].m_pszHeader,
				p, cb, pReq->m_bSendBody, pReq->m_bKeepAlive);
			return;
		}
	HttpError(nSock, 404, szNotFound);
}

//	HttpRequest
//	This processes the request block to see what is being asked of us.  We
//	break apart the first line into the pieces we need.  Return is true if
//	we expect another request.
static bool HttpRequest(SOCKET nSock, char * pBuff)
{	int i;
	char * pReq;
	struct EscWebRequest req;
	req.m_pBuff = sPageBuff;
	req.m_pszPath = NULL;
	req.m_pszParams = NULL;
	req.m_pszHeader = NULL;
	req.m_pszPost = NULL;
	req.m_nMajorVers = 0;
	req.m_nMinorVers = 9;	//	assume HTTP 0.9 spec
	req.m_bKeepAlive = false;
	req.m_bSendBody = true;
	//	find the method, request, HTTP version and header sections...
	for (i=0; pBuff[i]>' '; i++);
	if (pBuff[i] < ' ')
		return false;	//	didn't find the request
	pBuff[i++] = 0;
	pReq = pBuff + i;
	for (; pBuff[i]>' '; i++);
	if (pBuff[i] == ' ' && memcmp(pBuff+i+1, "HTTP/", 5) == 0)
	{	char * p;
		req.m_nMajorVers = strtol(pBuff+i+6, &p, 10);
		if (*p == '.')
			req.m_nMinorVers = strtol(p+1, NULL, 10);
		else
			req.m_nMinorVers = 0;
	}
	if (pBuff[i])
	{	//	see if there is a request body (post parameters)...
		for (req.m_pszPost=pBuff+i; *req.m_pszPost; req.m_pszPost++)
			if (memcmp(req.m_pszPost, szHeaderEnd, 4) == 0)
			{	*((char *)req.m_pszPost) = 0;
				req.m_pszPost += 4;
				break;
			}
		if (! *req.m_pszPost)
			req.m_pszPost = NULL;
		//	move ahead to the header (if any)
		pBuff[i++] = 0;
		while (pBuff[i] && pBuff[i] != '\r' && pBuff[i] != '\n')
			i++;
		while (pBuff[i] && (pBuff[i] == '\r' || pBuff[i] == '\n'))
			i++;
		if ((pBuff + i) != req.m_pszPost)
			req.m_pszHeader = pBuff + i;
	}
	//	check the request for parameters...
	for (i=0; pReq[i] && pReq[i] != '?'; i++);
	if (pReq[i] == '?')
	{	pReq[i++] = 0;
		req.m_pszParams = pReq + i;
	}
	req.m_pszPath = pReq;
	//	see if there is a keep-alive...
	if (req.m_pszHeader)
	{	for (i=0; req.m_pszHeader[i]; i++)
			if (_strnicmp(req.m_pszHeader+i, "Connection:", 11) == 0)
			{	i += 11;
				while (req.m_pszHeader[i] == ' ')
					i++;
				if (_strnicmp(req.m_pszHeader+i, "keep-alive", 10) == 0)
					req.m_bKeepAlive = true;
				break;
			}
	}
	if (strcmp(pBuff, "GET") == 0)
	{	req.m_pszPost = NULL;
		HttpDispatchRequest(nSock, &req);
	} else if (strcmp(pBuff, "HEAD") == 0)
	{	req.m_pszPost = NULL;
		req.m_bSendBody = false;
		HttpDispatchRequest(nSock, &req);
	} else if (strcmp(pBuff, "POST") == 0)
	{	if (! req.m_pszPost)
			req.m_pszPost = "";
		req.m_bKeepAlive = false;
		HttpDispatchRequest(nSock, &req);
	} else
	{	HttpError(nSock, 501, szNotImpl);
		req.m_bKeepAlive = false;
	}
	//	other methods can be added if needed
	return req.m_bKeepAlive;
}

//	CheckHttp
//	This checks the web server listen socket for a connection.  If one is
//	present, we read in the request.  This is not a particularly robust
//	implementation: the same thread does the listen and the receive, which
//	means that concurrent requests will block.
static void CheckHttp(void)
{	struct timeval tv;
	struct fd_set fdsRead;
	struct sockaddr s;
	int cbAddr;
	SOCKET nSockReq;
	int i, n, cb;
	DWORD dwTick;
	char reqbuff[HTTP_BUFF_SIZE];

	//	see if there is a waiting connection...
	tv.tv_sec = tv.tv_usec = 0;
	FD_ZERO(&fdsRead);
	FD_SET(nSockWeb, &fdsRead);
	n = selectsocket(nSockWeb+1, &fdsRead, NULL, NULL, &tv);
	if (! n || n == SOCKET_ERROR)
	{	if (n == SOCKET_ERROR)
			dprintf(PRINTF_ERROR, "CheckHttp failed selectsocket\n");
		return;
	}
	cbAddr = sizeof(s);
	nSockReq = accept(nSockWeb, (struct sockaddr*)&s, &cbAddr);
	if (nSockReq == INVALID_SOCKET)
	{	dprintf(PRINTF_ERROR, "CheckHttp failed accept\n");
		return;
	}
alive:
	dwTick = nTickCount;	//	volatile nTickCount updated by interrupt
	cb = 0;
	while (1)
	{	FD_ZERO(&fdsRead);
		FD_SET(nSockReq, &fdsRead);
		n = selectsocket(nSockReq+1, &fdsRead, NULL, NULL, &tv);
		if (! (n))
		{	DWORD nt = nTickCount;
			if ((nt - dwTick) < 100)
			{	//	if we haven't timed out yet, keep trying...
				SysYield();
				continue;
			}
		}
		if (!n || n == SOCKET_ERROR)
			break;
		n = recv(nSockReq, reqbuff+cb, HTTP_BUFF_SIZE-1-cb, 0);
		if (n == SOCKET_ERROR)
		{	n = GetLastError(nSockReq);
			if (n != EWOULDBLOCK && n != ECONNRESET)
			{	bool b;
				dprintf(PRINTF_ERROR|PRINTF_FLASH, GetErrorMessage(n, &b));
				dprintf(PRINTF_ERROR|PRINTF_FLASH, " - CheckHttp recv\n");
			}
			n = 0;
			cb = 0;
			break;
		}
		cb += n;
		if (cb >= (HTTP_BUFF_SIZE-1))
			break;	//	buffer full!
		//	check for exit conditions...
		if (!n)
			break;
		for (i=0; i<(cb-4); i++)
			if (memcmp(reqbuff+i, szHeaderEnd, 4) == 0)
				break;
		if (i < (cb-4))
		{	n = i;
			for (i=0; i<(n-17); i++)
				if (_strnicmp(reqbuff+i, "\nContent-Length: ", 17) == 0)
					break;
			if (i == (n-17))
				break;	//	we have full header and no body
			n += (strtol(reqbuff+i+17, NULL, 10) + 4);
			if (cb >= n)
				break;	//	we have at least as many bytes as expected
		}
	}
	if (n == SOCKET_ERROR)
	{	n = GetLastError(nSockReq);
		if (n != EWOULDBLOCK && n != ECONNRESET)
		{	bool b;
			dprintf(PRINTF_ERROR|PRINTF_FLASH, GetErrorMessage(n, &b));
			dprintf(PRINTF_ERROR|PRINTF_FLASH, " - CheckHttp selectsocket\n");
		}
		cb = 0;
	}
	if (cb)
	{	reqbuff[cb] = 0;
		if (HttpRequest(nSockReq, reqbuff))
		{	dprintf(PRINTF_DEBUG, "Going for more on keep-alive\n");
			goto alive;
		}
		SysSleep(2);	//	wait for a short period before closing the socket
	}
	shutdown(nSockReq, 0);
	closesocket(nSockReq);
}

//	EscWebServerThreadProc
//	This is simply the initialization and main loop for running our
//	example web server.
void EscWebServerThreadProc(void)
{	if ((nSockWeb = socket(PF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
	{	dprintf(PRINTF_ERROR|PRINTF_FLASH, "Unable to create web server "
			"socket");
		return;
	}
	if (BindSocketPort(nSockWeb, HTTP_PORT_NUMBER) && ListenSock(nSockWeb))
	{	while (! bKillWebThread)
		{	CheckHttp();
			SysYield();
		}
	}
	closesocket(nSockWeb);
}
